//	cria - DCC plugin
//	Copyright (c) 2002-2005 Brand Huntsman and Leif Ames
//


#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>		// gettimeofday()
#include <sys/stat.h>		// stat()

#include "common.h"
#include "functions.h"

#ifndef bzero
  #include <strings.h>		// bzero()
#endif

#ifndef __socklen_t_defined
  typedef unsigned int socklen_t;
#endif

#define DCC_BUFSIZE 4096
#define DCC_MAX_FILENAME 200
#define DCC_MAX_DIRNAME 500
#define DCC_MAX_PATHNAME DCC_MAX_DIRNAME + DCC_MAX_FILENAME


//////////////////////////////////////////////////////////////////////////

// local
s_plug *dcc_plug;
s_page *p_dcc;
char *dcc_get_dir;

typedef struct s_dcc_get {
	s_network *network;
	char *from;
	uint port;
	int downloaded, size;
	char *filename;
	int socket;
	FILE *fp;
	double start_time;
	struct s_dcc_get *next;
} s_dcc_get;

typedef struct s_dcc_send {
	s_network *network;
	char *to;
	int uploaded, size;
	char *filename;
	int socket;
	FILE *fp;
	double start_time;
	struct s_dcc_send *next;
} s_dcc_send;

struct s_dcc_get *dg_head;
struct s_dcc_send *ds_head;
char *buffer;

////////////////////////////////////////////////////////////////////////// PRIVATE


int intToIP( char *IP, uint address ){
	unsigned char a, b, c, d;

	d = address & 255;
	address >>= 8;
	c = address & 255;
	address >>= 8;
	b = address & 255;
	address >>= 8;
	a = address & 255;
	snprintf(IP, 16, "%i.%i.%i.%i", a, b, c, d);
	return(0);
}


int dcc_timer_callback( uint id );


////////////////////


int dg_recv_packet( int socket, FILE *fp ){
	// returns -1 if failed
	int count;

	count = recv(socket, buffer, DCC_BUFSIZE, 0);
	if(count <= 0 || count > DCC_BUFSIZE) return(-1);

	if(fwrite(buffer, 1, count, fp) != count); // warn_unused_result
	fflush(fp);
	return(count);
}
int dg_send_ack( int socket, uint size ){
	// returns -1 if failed
	int count;
	char ack[4];

	if(socket){
		*(uint *)ack = htonl(size); // warning: dereferencing type-punned pointer will break strict-aliasing rules
		count = send(socket, ack, 4, 0);
		if(count != 4) return(-1);

		return(count);
	}
	return(-1);
}


s_dcc_get *dg_find( int socket ){
	s_dcc_get *dg;

	for(dg = dg_head; dg != NULL; dg = dg->next)
		if(dg->socket == socket)
			return(dg);
	return(NULL);
}


s_dcc_get *dg_create( ){
	s_dcc_get *dg;

	dg = (s_dcc_get *)track_malloc(sizeof(struct s_dcc_get), "DCC get node");
	if(dg != NULL){
		if(dg_head == NULL && ds_head == NULL){
			buffer = (char *)track_malloc(DCC_BUFSIZE, "DCC buffer");
		}

		dg->next = dg_head;
		dg_head = dg;

		dg->socket = 0;
		dg->from = NULL;
		dg->filename = NULL;
		dg->fp = NULL;

		if(p_dcc == NULL){
			p_dcc = pageCreate(NULL, "!dcc", NULL);

			// enable refresh timer
			plugHookTimer(dcc_plug, 0, 2, &dcc_timer_callback);
		}
	}
	return(dg);
}


void dg_free( s_dcc_get *dg ){
	s_dcc_get *dg_tmp;

	plugRemoveHookSocket(dcc_plug, dg->socket);
	if(dg->socket) close(dg->socket);
	if(dg->fp) fclose(dg->fp);
	if(dg->from) track_free(dg->from, "PLUG dcc from");
	if(dg->filename) track_free(dg->filename, "PLUG dcc filename");

	// remove dg node from dg_head list
	if(dg == dg_head){
		dg_head = dg->next;
	} else {
		for(dg_tmp = dg_head; dg_tmp->next != dg; dg_tmp = dg_tmp->next);
		dg_tmp->next = dg->next;
	}

	if(dg_head == NULL && ds_head == NULL){
		track_free(buffer, "PLUG dcc buffer");

		//disable refresh timer
		plugRemoveHookTimer(dcc_plug, 0);

		pageFree(p_dcc);
		p_dcc = NULL;
	}

	track_free(dg, "PLUG dcc get entry");
}


int dg_verify_filename( char *dirname, char *filename ){
	char tmppath[DCC_MAX_PATHNAME], tmpfile[DCC_MAX_FILENAME + 50];
	FILE *fp;

	if(strlen(filename) >= DCC_MAX_FILENAME + 49) return(0);

	snprintf(tmppath, DCC_MAX_PATHNAME, "%s/%s", dirname, filename);
	if((fp = fopen(tmppath, "r")) != NULL){
		fclose(fp);
		snprintf(tmpfile, DCC_MAX_FILENAME + 50, "_%s", filename);
		strncpy(filename, tmpfile, DCC_MAX_FILENAME);
		return(dg_verify_filename(dirname, filename));
	}
	return(1);
}


int dg_socket_callback( int socket ){
	s_dcc_get *dg;
	int n;
	double end_time, speed, size;
	char str_speed[20], str_size[20];
	struct timeval tv;

	dg = dg_find(socket);

	// get packet
	n = dg_recv_packet(socket, dg->fp);
	if(n == -1){
		msgServer(0, dg->network, MFP_DCC_LOSTCON, 3, dg->filename, dg->from, "GET");
		dg_free(dg);
		return(0);
	}
	dg->downloaded += n;

	// send received count back to sender
	if(dg_send_ack(socket, dg->downloaded) == -1){
		msgServer(0, dg->network, MFP_DCC_LOSTCON, 3, dg->filename, dg->from, "GET");
		dg_free(dg);
		return(0);
	}

	if(dg->downloaded == dg->size){
		gettimeofday(&tv, NULL);
		end_time = (double)tv.tv_sec + (double)tv.tv_usec/1000000;

		speed = dg->size / (1024 * (end_time - dg->start_time));
		if(speed >= 1.0)
			snprintf(str_speed, sizeof(str_speed), "%.1f", speed);
		else
			snprintf(str_speed, sizeof(str_speed), "%.3f", speed);

		size = ((float)dg->size) / 1024;
		if(size >= 1.0)
			snprintf(str_size, sizeof(str_size), "%.1f", size);
		else
			snprintf(str_size, sizeof(str_size), "%.3f", size);

		msgServer(0, dg->network, MFP_DCC_GET_COMPLETE, 4,
			str_speed, str_size, dg->filename, dg->from);
		dg_free(dg);
	}

	return(0);
}


////////////////////


int ds_send_packet( int socket, FILE *fp ){
	// returns -1 if failed
	int count, bytes;

	if(socket){
		bytes = fread(buffer, 1, DCC_BUFSIZE, fp);
		count = send(socket, buffer, bytes, 0);
		return(count);
	}
	return(-1);
}
int ds_get_ack( int socket, uint *size ){
	// returns -1 if failed
	int count;
	char ack[4];

	if(socket){
		count = recv(socket, ack, 4, 0);
		if(count != 4) return(-1);
		*size = ntohl(*(uint *)ack); // warning: dereferencing type-punned pointer will break strict-aliasing rules
		return(count);
	}
	return(-1);
}


s_dcc_send *ds_find( int socket ){
	s_dcc_send *ds;

	for(ds = ds_head; ds != NULL; ds = ds->next)
		if(ds->socket == socket)
			return(ds);
	return(NULL);
}


s_dcc_send *ds_create( ){
	s_dcc_send *ds;

	ds = (s_dcc_send *)track_malloc(sizeof(struct s_dcc_send), "DCC send node");
	if(ds != NULL){
		if(ds_head == NULL && dg_head == NULL){
			buffer = (char *)track_malloc(DCC_BUFSIZE, "DCC buffer");
		}

		ds->next = ds_head;
		ds_head = ds;

		ds->socket = 0;
		ds->to = NULL;
		ds->filename = NULL;
		ds->fp = NULL;

		if(p_dcc == NULL){
			p_dcc = pageCreate(NULL, "!dcc", NULL);

			// enable refresh timer
			plugHookTimer(dcc_plug, 0, 2, &dcc_timer_callback);
		}
	}
	return(ds);
}


void ds_free( s_dcc_send *ds ){
	s_dcc_send *ds_tmp;

	plugRemoveHookSocket(dcc_plug, ds->socket);
	if(ds->socket) close(ds->socket);
	if(ds->fp) fclose(ds->fp);
	if(ds->to) track_free(ds->to, "PLUG dcc to");
	if(ds->filename) track_free(ds->filename, "PLUG dcc filename");

	// remove ds node from ds_head list
	if(ds == ds_head){
		ds_head = ds->next;
	} else {
		for(ds_tmp = ds_head; ds_tmp->next != ds; ds_tmp = ds_tmp->next);
		ds_tmp->next = ds->next;
	}

	if(ds_head == NULL && dg_head == NULL){
		track_free(buffer, "PLUG dcc buffer");

		//disable refresh timer
		plugRemoveHookTimer(dcc_plug, 0);

		pageFree(p_dcc);
		p_dcc = NULL;
	}

	track_free(ds, "PLUG dcc send entry");
}


int ds_socket_callback( int socket ){
	s_dcc_send *ds;
	int count;
	double end_time, speed, size;
	char str_speed[20], str_size[20];
	struct timeval tv;

	ds = ds_find(socket);

	if(ds_get_ack(socket, (uint *)&count) == -1){
		msgServer(0, ds->network, MFP_DCC_LOSTCON, 3, ds->filename, ds->to, "SEND");
		ds_free(ds);
		return(0);
	}
	if(count == ds->uploaded){
		if(count == ds->size){
			gettimeofday(&tv, NULL);
			end_time = (double)tv.tv_sec + (double)tv.tv_usec/1000000;

			speed = ds->size / (1024 * (end_time - ds->start_time));
			if(speed >= 1.0)
				snprintf(str_speed, 20, "%.1f", speed);
			else
				snprintf(str_speed, 20, "%.3f", speed);

			size = ((float)ds->size) / 1024;
			if(size >= 1.0)
				snprintf(str_size, 20, "%.1f", size);
			else
				snprintf(str_size, 20, "%.3f", size);

			msgServer(0, ds->network, MFP_DCC_SEND_COMPLETE, 4,
				str_speed, str_size, ds->filename, ds->to);
			ds_free(ds);
		} else {
			if((count = ds_send_packet(socket, ds->fp)) == -1){
				msgServer(0, ds->network, MFP_DCC_LOSTCON, 3, ds->filename, ds->to, "SEND");
				ds_free(ds);
				return(0);
			} else {
				ds->uploaded += count;
			}
		}
	}

	return(0);
}


int ds_accept_socket_callback( int socket ){
	s_dcc_send *ds;
	struct sockaddr_in remote;
	socklen_t addr_len;
	struct timeval tv;

	ds = ds_find(socket);

	addr_len = sizeof(struct sockaddr_in);
	if((ds->socket = accept(socket, (struct sockaddr *)&remote, &addr_len)) < 0){
		msgServer(3, ds->network, MFP_DCC_NOSOCK, 3, ds->filename, ds->to, "SEND");
		ds_free(ds);
		return(0);
	}

	plugRemoveHookSocket(dcc_plug, socket);
	plugHookSocket(dcc_plug, ds->socket, &ds_socket_callback);
	close(socket);

	gettimeofday(&tv, NULL);
	ds->start_time = (double)tv.tv_sec + (double)tv.tv_usec/1000000;

	if((ds->uploaded = ds_send_packet(ds->socket, ds->fp)) == -1){
		msgServer(0, ds->network, MFP_DCC_LOSTCON, 3, ds->filename, ds->to, "SEND");
		ds_free(ds);
		return(0);
	}

	return(0);
}


void ds_create_server( int type, s_network *network, char *nick, char *file ){
	uint address, port;
	struct stat stat_buffer;
	struct sockaddr_in local;
	struct in_addr inp;
	socklen_t addr_len;
	s_dcc_send *ds;
	int s;
	char *filename, size_str[20], path[LEN_PATHNAME];

	if((s = socket(AF_INET, SOCK_STREAM, IPPROTO_IP)) < 0){
		msgServer(3, network, MFP_DCC_NOSOCK, 3, file, nick, "SEND");
		return;
	}

	memset(&local, 0, sizeof(struct sockaddr_in));
	local.sin_family = AF_INET;
	local.sin_port = htons(0);
	local.sin_addr.s_addr = htonl(INADDR_ANY);
	#ifdef __CHAR_BZERO__
		bzero((char *)&(local.sin_zero), 8);
	#else
		bzero(&(local.sin_zero), 8);
	#endif

	if(bind(s, (struct sockaddr *)&local, sizeof(struct sockaddr_in)) == -1){
		msgServer(3, network, MFP_DCC_NOSOCK, 3, file, nick, "SEND");
		close(s);
		return;
	}

	addr_len = sizeof(struct sockaddr_in);
	getsockname(s, (struct sockaddr *)&local, &addr_len);
	port = ntohs(local.sin_port);

	if(var_ip_addr == NULL)
		inet_aton(network->local_ip_addr, &inp);
	else inet_aton(var_ip_addr, &inp);
	address = ntohl(inp.s_addr);

	if(type == 0){									// SEND
		ds = ds_create();

		ds->socket = s;

		if(*file == '~')
			snprintf(path, LEN_PATHNAME, "%s%s", home, file+1);
		else
			snprintf(path, LEN_PATHNAME, "%s", file);

		if((ds->fp = fopen(path, "r")) == NULL){
			msgServer(3, network, MFP_DCC_NOFILE, 3, file, nick, "SEND");
			ds_free(ds);
			return;
		}

		ds->network = network;
		ds->uploaded = 0;
		stat((const char *)path, &stat_buffer);
		ds->size = (int)stat_buffer.st_size;
		setMakeStr(&ds->to, nick);
		setMakeStr(&ds->filename, file);

		// trim off the path
		for(filename = file + strlen(file); filename != file && *(filename-1) != '/'; filename--);

		netSend(network, "PRIVMSG %s :\001DCC SEND %s %d %d %d\001\n",
			nick, filename, address, port, (int)stat_buffer.st_size);

		snprintf(size_str, 20, "%d", (int)stat_buffer.st_size);
		msgPage(p_dcc, 3, network, MFP_DCC_SEND_STATUS, 5,
			"0", "0", size_str, path, nick);

		ds->start_time = 0.0;
	} else {									// CHAT

  // TODO: create chat node

		netSend(network, "PRIVMSG %s :\001DCC CHAT chat %d %d\001\n",
			nick, address, port);

msgServer(3, network, MF_DEBUG, 1, "sent unsupported dcc chat request");

	}

	listen(s, 1);
	plugHookSocket(dcc_plug, s, &ds_accept_socket_callback);
}


////////////////////


void dsg_close( s_network *network, char *nick ){
	s_dcc_send *ds, *nds;
	s_dcc_get *dg, *ndg;

	ds = ds_head;
	while(ds != NULL){
		nds = ds->next;
		if(ds->network == network)
			if(!strcmp(ds->to, nick))
				ds_free(ds);

		ds = nds;
	}

	dg = dg_head;
	while(dg != NULL){
		ndg = dg->next;
		if(dg->network == network)
			if(!strcmp(dg->from, nick))
				dg_free(dg);

		dg = ndg;
	}
}


////////////////////


int dcc_timer_callback( uint id ){
	s_dcc_send *ds;
	s_dcc_get *dg;
	char str_speed[20], str_size[20], str_percentage[4];
	double current_time, speed, size;
	struct timeval tv;

	p_dcc->new_messages = 0;
	bufferClear(p_dcc);
	if(p_current == p_dcc) msgDisplayPage(p_dcc);

	for(ds = ds_head; ds != NULL; ds = ds->next){
		gettimeofday(&tv, NULL);
		current_time = (double)tv.tv_sec + (double)tv.tv_usec/1000000;

		snprintf(str_percentage, 4, "%d", (int)(((float)ds->uploaded / (float)ds->size) * 100.0));

		size = ((float)ds->size) / 1024;
		if(size >= 1.0)
			snprintf(str_size, 20, "%.1f", size);
		else
			snprintf(str_size, 20, "%.3f", size);

		if(ds->start_time > 0.0){
			speed = ds->uploaded / (1024 * (current_time - ds->start_time));
			if(speed >= 1.0)
				snprintf(str_speed, 20, "%.1f", speed);
			else
				snprintf(str_speed, 20, "%.3f", speed);

			msgPage(p_dcc, 3, ds->network, MFP_DCC_SEND_STATUS, 5,
				str_percentage, str_speed, str_size, ds->filename, ds->to);
		} else {
			msgPage(p_dcc, 3, ds->network, MFP_DCC_SEND_STATUS, 5,
				str_percentage, "---", str_size, ds->filename, ds->to);
		}
	}

	for(dg = dg_head; dg != NULL; dg = dg->next){
		gettimeofday(&tv, NULL);
		current_time = (double)tv.tv_sec + (double)tv.tv_usec/1000000;

		snprintf(str_percentage, 4, "%d", (int)(((float)dg->downloaded / (float)dg->size) * 100.0));

		speed = dg->downloaded / (1024 * (current_time - dg->start_time));
		if(speed >= 1.0)
			snprintf(str_speed, 20, "%.1f", speed);
		else
			snprintf(str_speed, 20, "%.3f", speed);

		size = ((float)dg->size) / 1024;
		if(size >= 1.0)
			snprintf(str_size, 20, "%.1f", size);
		else
			snprintf(str_size, 20, "%.3f", size);

		msgPage(p_dcc, 3, dg->network, MFP_DCC_GET_STATUS, 5,
			str_percentage, str_speed, str_size, dg->filename, dg->from);
	}

	return(0);	// 0 = repeat timer
}


static int privmsg_callback( s_network *network, char *from, char *data, int datalen ){
	char *message, *host, *type_str, *arg_str, *address_str, *port_str, *size_str, *filename;
	char pathname[DCC_MAX_PATHNAME], dirname[DCC_MAX_DIRNAME], temp_size[20], addr_ip[16];
	double size;
	uint src;
	struct timeval tv;
	s_dcc_get *dg;

	message = skipword(data) + 1;
	if(*message == '\001'){
		if(strncmp(message+1, "DCC", 3)){
			*(message-2) = ' ';
			return(1);
		}
	} else {
		*(message-2) = ' ';
		return(1);
	}

  // data = destination (channel or you)

	host = skipnick(from);

	*(message + strlen(message) - 1) = '\0';	// strip ctcp end character
	type_str = skipword(message);
	arg_str = skipword(type_str);

	if(!strcmp(type_str, "ERROR")){
		msgServer(3, network, MFP_DCC_ERROR, 2, arg_str, from);
		return(0);
	} else if(!strcmp(type_str, "SEND")){
		address_str = skipword(arg_str);
		port_str = skipword(address_str);
		size_str = skipword(port_str);

		if(proxy){
			char msg[500];
			snprintf(msg, 500, "--- Can't receive DCC files over proxy (from %s)", from);
msgServer(3, NULL, MF_DEBUG, 1, msg);
			netSend(network, "PRIVMSG %s :\001DCC ERROR I'm behind a proxy and can't receive files\001\n", from);
			return(0);
		}

		dg = dg_create();

		dg->network = network;
		setMakeStr(&dg->from, from);
		intToIP(addr_ip, cria_atoui(address_str));
		dg->port = cria_atoui(port_str);
		dg->downloaded = 0;
		dg->size = cria_atoui(size_str);

		if(arg_str == NULL || address_str == NULL || dg->port == 0 || dg->size == 0){
			char msg[500];
			snprintf(msg, 500, "--- Received malformed DCC SEND header (from %s), aborting", from);
msgServer(3, network, MF_DEBUG, 1, msg);
			netSend(network, "PRIVMSG %s :\001DCC ERROR I can't process your malformed DCC SEND header\001\n", from);
			dg_free(dg);
			return(0);
		}

		// strip path from filename
		for(filename = arg_str + strlen(arg_str); filename != arg_str && *(filename-1) != '/'; filename--);

		if(strlen(filename) >= DCC_MAX_FILENAME){
			// shorten filename
			filename += strlen(filename) - DCC_MAX_FILENAME + 1;
		}

		// build directory path
		if(dcc_get_dir == NULL){
			*dirname = '\0';
		} else if(*dcc_get_dir == '~'){
			snprintf(dirname, DCC_MAX_DIRNAME, "%s%s", home, dcc_get_dir+1);
		} else strncpy(dirname, dcc_get_dir, DCC_MAX_DIRNAME);

		if(dg_verify_filename(dirname, filename) == 0){
			char msg[500];
			snprintf(msg, 500, "--- Can't rename incoming file \"%s\" (from %s), aborting", filename, from);
msgServer(3, network, MF_DEBUG, 1, msg);
			netSend(network,
				"PRIVMSG %s :\001DCC ERROR I already have a file named %s and can't rename the file you're trying to send (the filename is probably too long)\001\n",
				filename, from);
			dg_free(dg);
			return(0);
		}
		setMakeStr(&dg->filename, filename);

  // TODO: the following code can be used as a Denial of Service attack on cria.
  //		the client will lockup if the other end doesnt respond

		netGetHost(&src, addr_ip);
		if((dg->socket = netOpen(src, dg->port)) == 0){
			msgServer(3, network, MFP_DCC_NOSOCK, 3, filename, from, "GET");
			dg_free(dg);
			return(0);
		}

		snprintf(pathname, DCC_MAX_PATHNAME, "%s/%s", dirname, filename);
		if((dg->fp = fopen(pathname, "w")) == NULL){
			msgServer(3, network, MFP_DCC_NOFILE, 3, pathname, from, "GET");
			dg_free(dg);
			return(0);
		}

		size = ((float)dg->size) / 1024;
		if(size >= 1.0)
			snprintf(temp_size, 20, "%.1f", size);
		else
			snprintf(temp_size, 20, "%.3f", size);
		msgPage(p_dcc, 3, network, MFP_DCC_GET_STATUS, 5,
			"0", "0", temp_size, filename, from);

		gettimeofday(&tv, NULL);
		dg->start_time = (double)tv.tv_sec + (double)tv.tv_usec/1000000;

		plugHookSocket(dcc_plug, dg->socket, &dg_socket_callback);
	} else if(!strcmp(type_str, "CHAT")){
//		address_str = skipword(arg_str);
//		port_str = skipword(address_str);
//		size_str = skipword(port_str);

  // TODO: add support

		msgServer(3, network, MFP_DCC_UNKNOWN, 3, arg_str, type_str, from);

	} else {
		msgServer(3, network, MFP_DCC_UNKNOWN, 3, arg_str, type_str, from);
	}

	return(0);
}


int dcc_prepare_parameters( int uses_file, char **net, char **nick, char **file, s_network **network ){
	if(uses_file == 1){
		// send
		if(**file == '\0'){
			*file = *nick;
			*nick = *net;
			if(**file == '\0'){
				msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/DCC SEND [network] nick file");
				return(1); // exit
			}
			*network = n_current;
		}
	} else if(uses_file == -1){
		// get
//		if(**file == '\0'){
//			*file = *nick;
//			*nick = *net;
//			if(**file == '\0'){
//				msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/DCC GET [network] nick [file]");
//				return(1); // exit
//			}
//			*network = n_current;
//		}

  // TODO: not sure how to handle both optional parameters

	} else {
		// chat and close
		if(**nick == '\0'){
			*nick = *net;
			*network = n_current;
		}
	}

	if(*network == NULL){
		if((*network = networkGet(*net)) == NULL){
			msgCurrentPage(-1, NULL, MF_ERROR, 1, "Unknown network");
			return(1); // exit
		}
	}
	if((*network)->socket == 0){
		msgCurrentPage(-1, NULL, MF_ERROR, 1, "You must be connected to the IRC server to use DCC");
		return(1); // exit
	}

	return(0);
}
int dcc_callback( char *data, int datalen ){
	// DCC SEND [network] nick file
	// DCC GET [network] nick [file]
	// DCC CHAT [network] nick
	// DCC CLOSE [network] nick
	char *nick, *file, *net;
	s_network *network = NULL;

	net = skipword(data);
	nick = skipword(net);
	file = skipword(nick);

	if(stringcompare(data, "send")){
		if(dcc_prepare_parameters(1, &net, &nick, &file, &network)) return(0); // dont passthru
		if(*nick == '#' || *file == '\0'){
			msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/DCC SEND [network] nick file");
		} else {
			if(proxy){
				msgCurrentPage(-1, NULL, MF_ERROR, 1, "Can't use DCC SEND over proxy");
				return(0); // dont passthru
			}

			ds_create_server(0, network, nick, file);
		}
	} else if(stringcompare(data, "chat")){
		if(dcc_prepare_parameters(0, &net, &nick, &file, &network)) return(0); // dont passthru
		if(*nick == '#' || *file != '\0'){
			msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/DCC CHAT [network] nick");
		} else {
			if(proxy){
				msgCurrentPage(-1, NULL, MF_ERROR, 1, "Can't use DCC CHAT over proxy");
				return(0); // dont passthru
			}

			ds_create_server(1, network, nick, NULL);
		}
	} else if(stringcompare(data, "get")){
		if(dcc_prepare_parameters(-1, &net, &nick, &file, &network)) return(0); // dont passthru

		// currently uses auto-get
		msgCurrentPage(-1, NULL, MF_FINISH_ME, 1, "DCC GET [network] nick [file]");

	} else if(stringcompare(data, "close")){
		if(dcc_prepare_parameters(0, &net, &nick, &file, &network)) return(0); // dont passthru
		if(*nick == '#' || *file != '\0'){
			msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/DCC CLOSE [network] nick");
		} else dsg_close(network, nick);
	} else {
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/DCC SEND [network] nick file");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/DCC GET [network] nick [file]");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/DCC CHAT [network] nick");
		msgCurrentPage(-1, NULL, MF_SYNTAX, 1, "/DCC CLOSE [network] nick");
	}

	return(0); // dont passthru
}


////////////////////////////////////////////////////////////////////////// GLOBAL


int dccInit( s_plug *self ){
	if(!plugVersion(0,2,8,4)){
		msgServer(3, NULL, MF_PLUG_REQUIRES, 2, "0.2.8cvs4", "DCC");
		return(0);
	}

	dcc_plug = self;
	p_dcc = NULL;
	dg_head = NULL;
	ds_head = NULL;
	plugHookUser(self, "dcc", 0, &dcc_callback);
	plugHookServer(self, "PRIVMSG", &privmsg_callback);

	dcc_get_dir = NULL;
	plugHookSet(self, "dcc_get_dir", SET_STR, (void *)&dcc_get_dir);
	setMakeStr(&dcc_get_dir, home);

	return(1);
}


void dccUnload( ){
	s_dcc_get *dg, *dg_tmp;
	s_dcc_send *ds, *ds_tmp;

	dg = dg_head;
	while(dg != NULL){
		dg_tmp = dg->next;
		dg_free(dg);
		dg = dg_tmp;
	}

	ds = ds_head;
	while(ds != NULL){
		ds_tmp = ds->next;
		ds_free(ds);
		ds = ds_tmp;
	}
}
